@@ -173,6 +173,15 @@ struct config {
     bool box_drawings_uses_font_glyphs;
     bool can_shape_grapheme;
 
+    struct {
+        enum {
+            OSC52_DISABLED,
+            OSC52_COPY_ENABLED,
+            OSC52_PASTE_ENABLED,
+            OSC52_ENABLED,
+        } osc52;
+    } security;
+
     struct {
         bool urgent;
         bool notify;
